// simple FFT of mono audio data using libsndfile and libfftw3

// usage:
// gcc -std=c99 -Wall -Wextra -pedantic -O3 -o fft fft.c -lfftw3 -lsndfile -lm
// ./fft 20 pink-noise-src.wav > pink-noise-src.dat
// ./fft 20 pink-noise-mic.wav > pink-noise-mic.dat

// more info:
// <http://mathr.co.uk/blog/2015-07-15_modes_in_a_box.html>

#include <complex.h>
#include <math.h>
#include <stdlib.h>
#include <fftw3.h>
#include <sndfile.h>

// midi note to frequency
static inline double mtof(double f) {
  return 8.17579891564 * exp(0.0577622650 * fmin(f, 1499));
}

// frequency to midi note
static inline double ftom(double f) {
  return 17.3123405046 * log(0.12231220585 * f);
}

int main(int argc, char **argv) {

  // FFT size
  if (argc != 3) { return 1; }
  int length2 = atoi(argv[1]);
  if (! (4 <= length2 && length2 <= 24)) { return 1; }
  int frames = 1 << length2;

  // prepare FFT
  double *data = fftw_malloc(frames * sizeof(*data));
  if (! data) { return 1; }
  complex double *fft = fftw_malloc(frames * sizeof(*fft));
  if (! fft) { return 1; }
  fftw_plan plan = fftw_plan_dft_r2c_1d(frames, data, fft, FFTW_ESTIMATE);

  // read input
  SF_INFO info;
  info.format = 0;
  SNDFILE *in = sf_open(argv[2], SFM_READ, &info);
  if (! in) { return 1; }
  if (info.channels != 1) { return 1; }
  if (info.frames < frames) { return 1; }
  if (frames != sf_read_double(in, data, frames)) { return 1; }
  double SR = info.samplerate;

  // window input
  for (int i = 0; i < frames; ++i) {
    data[i] *= 1 - cos(6.283185307179586 * (i + 0.5) / frames);
  }

  // perform fft
  fftw_execute(plan);

  // output spectrum
  int lo = floor(mtof(12) * frames / SR);
  int hi = ceil (mtof(72) * frames / SR);
  for (int i = lo; i <= hi; ++i) {
    double note = ftom(i * SR / frames);
    double magnitude = cabs(fft[i]);
    double phase = carg(fft[i]);
    printf("%.7e %.7e %.7e\n", note, magnitude, phase);
  }

  // done
  return 0;

}
